import json

from fastapi import Depends, Header, HTTPException, WebSocket, status
from fastapi.security import OAuth2PasswordBearer
from jose import jwt  # type: ignore
from pydantic import ValidationError

from .models.permission import Permission
from .models.token import TokenPayload
from .settings import settings

ALGORITHM = "HS256"

reusable_oauth2 = OAuth2PasswordBearer(tokenUrl=settings.OAUTH_TOKEN_URL)


def extract_permission(token: str) -> Permission:
    payload = jwt.decode(token, settings.SECRET_KEY, algorithms=[ALGORITHM])
    token_data = TokenPayload(**payload)
    permission_dict = json.loads(token_data.sub)
    return Permission(**permission_dict)


def get_current_permission(
    token: str = Depends(reusable_oauth2),
) -> Permission:
    try:
        return extract_permission(token)
    except (jwt.JWTError, ValidationError):
        raise HTTPException(
            status_code=status.HTTP_403_FORBIDDEN,
            detail="Could not validate credentials",
        )


def read_access(
    permission: Permission = Depends(get_current_permission),
) -> bool:
    if (
        permission.read_access is not True
        and permission.write_access is not True
    ):
        raise HTTPException(status_code=403, detail="Inadequeate permissions")
    return True


def write_access(
    permission: Permission = Depends(get_current_permission),
) -> bool:
    if permission.write_access is not True:
        raise HTTPException(status_code=403, detail="Inadequeate permissions")
    return True


# Browser cant connect with this method. Javascript doesn't
# allow headers (like Authorization) to be customized when connecting to a
# websocket. If browser support is needed in the future, may consider a
# ticket-based approach like the one detailed here:
# https://devcenter.heroku.com/articles/websocket-security#authentication-authorization  # noqa
async def get_current_permission_websocket(
    websocket: WebSocket, authorization: str = Header(None)
) -> Permission:
    try:
        assert authorization is not None
        splits = authorization.split(" ")
        assert len(splits) > 1
        token = splits[1]
        return extract_permission(token)
    except (jwt.JWTError, ValidationError, AssertionError):
        await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
        raise HTTPException(status_code=403, detail="Inadequeate permissions")


async def read_access_websocket(
    websocket: WebSocket,
    permission: Permission = Depends(get_current_permission_websocket),
) -> bool:
    if (
        permission.read_access is not True
        and permission.write_access is not True
    ):
        await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
        raise HTTPException(status_code=403, detail="Inadequeate permissions")
    return True


async def write_access_websocket(
    websocket: WebSocket,
    permission: Permission = Depends(get_current_permission_websocket),
) -> bool:
    if permission.write_access is not True:
        await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
        raise HTTPException(status_code=403, detail="Inadequeate permissions")
    return True



from typing import List

from pydantic import BaseModel

from .base_wrapper import BaseWrapper


class KernelMessage(BaseModel):
    message_id: str


class KernelMessageWrapper(BaseWrapper):
    kernel_message: KernelMessage


class KernelMessagesWrapper(BaseWrapper):
    kernel_messages: List[KernelMessage]



from ...utils import NotebookNode
from ..workdir_manager import WorkDirManager

raw_chdir_code = "cd {directory}"
raw_cwd_code = "pwd"


class ZshWorkDirManager(WorkDirManager):
    def get_chdir_code(self, directory: str) -> str:
        return f"{raw_chdir_code.format(directory=directory)}\n{raw_cwd_code}"

    def get_cwd_code(self) -> str:
        return raw_cwd_code

    def extract_cwd(self, response: NotebookNode) -> str:
        return response.text.strip()

    def parse_chdir_response(self, response: NotebookNode) -> str:
        return self.extract_cwd(response)

    def parse_cwd_response(self, response: NotebookNode) -> str:
        return self.extract_cwd(response)



import json
import typing
import uuid
from datetime import date, datetime

from starlette.responses import JSONResponse


def d1_pairs_hook(pairs):
    """ 
    To be used with json.load(s), attempts to convert any
    attribute with a name ending in 'uuid' into a UUID
    object.  The conversion may fail, in which case the
    value silently passes through.
    """
    ret_dict = {}
    for pair in pairs:
        (key, value) = pair
        if key.endswith("uuid"):
            try:
                value = uuid.UUID(value)
            except ValueError:
                pass
            except TypeError:
                pass
        ret_dict[key] = value
    return ret_dict


class D1Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.isoformat()
        if isinstance(obj, uuid.UUID):
            return str(obj)
        return json.JSONEncoder.default(self, obj)


class D1Response(JSONResponse):
    media_type = "application/json"

    def render(self, content: typing.Any) -> bytes:
        return json.dumps(
            content,
            ensure_ascii=False,
            allow_nan=False,
            indent=None,
            separators=(",", ":"),
            cls=D1Encoder,
        ).encode("utf-8")



from io import StringIO
from typing import Dict

from fastapi.testclient import TestClient
from pytest_mock import MockFixture


def test_convert_rmd_to_ipynb(
    client: TestClient, readonly_token_headers: Dict[str, str],
) -> None:
    with open("jupyter_d1/tests/notebooks/worldpop.Rmd", "rb") as f1:
        response = client.post(
            "/convert/rmd_to_ipynb",
            files={"file": ("worldpop.Rmd", f1)},
            headers=readonly_token_headers,
        )
        with open("jupyter_d1/tests/notebooks/worldpop.ipynb", "rb") as f2:
            assert response.content == f2.read()

    assert (
        response.headers["content-disposition"]
        == 'attachment; filename="worldpop.ipynb"'
    )


def test_convert_rmd_with_R_to_ipynb(
    client: TestClient, readonly_token_headers: Dict[str, str],
) -> None:
    with open("jupyter_d1/tests/notebooks/rmd_example.Rmd", "rb") as f1:
        response = client.post(
            "/convert/rmd_to_ipynb",
            files={"file": ("rmd_example.Rmd", f1)},
            headers=readonly_token_headers,
        )
        with open("jupyter_d1/tests/notebooks/rmd_example.ipynb", "rb") as f2:
            print(response.content)
            assert response.content == f2.read()

    assert (
        response.headers["content-disposition"]
        == 'attachment; filename="rmd_example.ipynb"'
    )


def test_convert_rmd_to_ipynb_bogus_file(
    client: TestClient, readonly_token_headers: Dict[str, str], mocker: MockFixture
) -> None:
    mock_jupytext = mocker.patch(
        "jupyter_d1.routers.convert.jupytext"
    )
    mock_jupytext.side_effect = KeyError("it did not work")

    response = client.post(
        "/convert/rmd_to_ipynb",
        files={"file": (
            "meta_info.json",
            StringIO("%%%5not valid file\n````{{yurp}}\nprint(g)\n```")
        )},
        headers=readonly_token_headers,
    )
    assert response.status_code == 400    
    assert (
        response.json()["detail"] == "Failed to convert Rmd file to ipynb, do you have "
        "an R kernel installed and added to the kernelspecs?"
    )



from abc import ABC, abstractmethod
from typing import List

from .vars_manager import VarsManager
from .workdir_manager import WorkDirManager


class KernelDefinition(ABC):
    @abstractmethod
    def create_vars_manager(self) -> VarsManager:
        raise NotImplementedError

    @abstractmethod
    def create_workdir_manager(self, workdir: str) -> WorkDirManager:
        raise NotImplementedError

    @property
    def kernel_options(self) -> List[str]:
        return []



import logging
import os
import shutil
import unittest
from pathlib import Path
from typing import AsyncGenerator, Dict, Generator

import pytest  # type: ignore
from async_asgi_testclient import (
    TestClient as WebSocketTestClient,  # type: ignore
)
from fastapi.logger import logger
from fastapi.testclient import TestClient

from jupyter_d1 import app
from jupyter_d1.settings import settings

from .utils import (
    get_permissionless_token_headers,
    get_readonly_token_headers,
    get_superuser_token_headers,
)

enable_logger_output = False
# enable_logger_output = True


@pytest.fixture()
def client() -> Generator:
    with TestClient(app) as c:
        yield c


@pytest.fixture()
def superuser_token_headers(client: TestClient) -> Dict[str, str]:
    return get_superuser_token_headers(client)


@pytest.fixture()
def readonly_token_headers(client: TestClient) -> Dict[str, str]:
    return get_readonly_token_headers()


@pytest.fixture()
def permissionless_token_headers(client: TestClient) -> Dict[str, str]:
    return get_permissionless_token_headers()


@pytest.fixture(scope="function")
def clear_notebooks(
    client: TestClient, superuser_token_headers: Dict[str, str]
) -> Generator:
    yield
    response = client.delete("/notebooks", headers=superuser_token_headers)
    assert response.status_code == 204
    response = client.get("/notebooks", headers=superuser_token_headers)
    assert response.status_code == 200


@pytest.fixture(scope="function")
def clear_notebook_directory(
    client: TestClient, superuser_token_headers: Dict[str, str]
) -> Generator:
    root_path = Path(settings.ROOT_DIR)
    root_path.mkdir(parents=True, exist_ok=True)
    yield
    if "/tmp" in str(root_path.absolute()):
        for filename in os.listdir(root_path):
            file_path = os.path.join(root_path, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                print("Failed to delete %s. Reason: %s" % (file_path, e))


@pytest.fixture(scope="function")
def clear_notebook_directory_2() -> Generator:
    workdir = Path("/tmp/test_nb_2/")
    workdir.mkdir(parents=True, exist_ok=True)
    yield
    for filename in os.listdir(workdir):
        file_path = os.path.join(workdir, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            elif os.path.isdir(file_path):
                shutil.rmtree(file_path)
        except Exception as e:
            print("Failed to delete %s. Reason: %s" % (file_path, e))


@pytest.fixture()
async def websocket_client() -> AsyncGenerator:
    async with WebSocketTestClient(app) as c:
        yield c


@pytest.fixture()
def ututils() -> unittest.TestCase:
    """Unittest TestCase so we can use useful things like assertCountEquals"""
    return unittest.TestCase()


def pytest_sessionstart(session):

    if enable_logger_output:
        # Print out logging (visible with `pytest -s`)
        logger.setLevel(logging.DEBUG)
        ch = logging.StreamHandler()
        ch.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
        )
        ch.setFormatter(formatter)
        logger.addHandler(ch)



import json
from typing import Dict

import pytest  # type: ignore
from fastapi.testclient import TestClient

from jupyter_d1.settings import settings

from .test_notebooks import upload_notebook


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestUndo:
    def get_uuids(self, cells):
        def uuid(cell):
            return cell["metadata"]["jupyter_d1"]["uuid"]

        return list(map(uuid, cells))

    def test_undo_delete(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers)

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        # Start with 4 cells
        assert len(cells) == 4
        cell = cells[2]
        cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]

        # Delete a cell
        response = client.delete(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert len(file_nb["cells"]) == 3

        # Now 3 cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 3

        # Undo deletion
        response = client.get(
            f"/notebooks/{uuid}/undo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert len(file_nb["cells"]) == 4

        # Back to original, 4 cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 4

        # Redo the deletion
        response = client.get(
            f"/notebooks/{uuid}/redo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert len(file_nb["cells"]) == 3

        # Back to 3 cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 3

    def test_undo_create(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers)

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        # Start with 4 cells
        assert len(cells) == 4
        cell = cells[2]
        cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]

        # Add a cell
        src = "__**Hello Callisto**__"
        params = {"before": cell_uuid, "cell_type": "markdown", "source": src}
        response = client.post(
            f"/notebooks/{uuid}/cells",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        new_cell = response.json()["cell"]
        assert new_cell["cell_type"] == "markdown"
        assert new_cell["source"] == "__**Hello Callisto**__"

        # Now 5 cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 5

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert len(file_nb["cells"]) == 5

        # Undo creation
        response = client.get(
            f"/notebooks/{uuid}/undo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert len(file_nb["cells"]) == 4

        # Back to original, 4 cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 4

        # Redo the addition
        response = client.get(
            f"/notebooks/{uuid}/redo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert len(file_nb["cells"]) == 5

        # Back to 5 cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 5

        # check the contents of the new cell are correct
        cell = cells[2]
        assert cell["cell_type"] == "markdown"
        assert cell["source"] == "__**Hello Callisto**__"
        # Same for file on disk
        assert file_nb["cells"][2]["cell_type"] == "markdown"
        assert file_nb["cells"][2]["source"] == ["__**Hello Callisto**__"]

    def test_undo_update(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers)

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        # Start with 4 cells
        assert len(cells) == 4
        cell = cells[2]
        cell_uuid = cell["metadata"]["jupyter_d1"]["uuid"]
        assert cell["cell_type"] == "code"
        assert cell["source"] == "2+5"

        # Update a cell
        src = "__**Hello Callisto**__"
        params = {"cell_type": "markdown", "source": src}
        response = client.patch(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 200
        cell = response.json()["cell"]
        assert cell["cell_type"] == "markdown"
        assert cell["source"] == src

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert file_nb["cells"][2]["source"] == [src]
        assert file_nb["cells"][2]["cell_type"] == "markdown"

        # Undo update
        response = client.get(
            f"/notebooks/{uuid}/undo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert file_nb["cells"][2]["source"] == ["2+5"]
        assert file_nb["cells"][2]["cell_type"] == "code"

        # Back to original content
        response = client.get(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            headers=superuser_token_headers,
        )
        cell = response.json()["cell"]
        assert response.status_code == 200
        assert cell["cell_type"] == "code"
        assert cell["source"] == "2+5"

        # Redo update
        response = client.get(
            f"/notebooks/{uuid}/redo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert file_nb["cells"][2]["source"] == [src]
        assert file_nb["cells"][2]["cell_type"] == "markdown"

        # Now updated content
        response = client.get(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            headers=superuser_token_headers,
        )
        cell = response.json()["cell"]
        assert response.status_code == 200
        assert cell["cell_type"] == "markdown"
        assert cell["source"] == src

    def test_undo_move(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers)

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        uuids = self.get_uuids(cells)
        assert len(uuids) == 4

        uuid_a = uuids[0]
        uuid_b = uuids[1]
        uuid_c = uuids[2]
        uuid_d = uuids[3]

        # 'move' with no query param should move uuid_b
        # to the end of the list
        response = client.get(
            f"/notebooks/{uuid}/cells/{uuid_b}/move",
            headers=superuser_token_headers,
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert self.get_uuids(file_nb["cells"]) == [
            uuid_a,
            uuid_c,
            uuid_d,
            uuid_b,
        ]

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        uuids = self.get_uuids(response.json()["cells"])
        assert uuids == [uuid_a, uuid_c, uuid_d, uuid_b]

        # Undo move
        response = client.get(
            f"/notebooks/{uuid}/undo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert self.get_uuids(file_nb["cells"]) == [
            uuid_a,
            uuid_b,
            uuid_c,
            uuid_d,
        ]

        # Back to original order
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        uuids = self.get_uuids(response.json()["cells"])
        assert uuids == [uuid_a, uuid_b, uuid_c, uuid_d]

        # Redo the move
        response = client.get(
            f"/notebooks/{uuid}/redo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert self.get_uuids(file_nb["cells"]) == [
            uuid_a,
            uuid_c,
            uuid_d,
            uuid_b,
        ]

        # Back to altered order
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        uuids = self.get_uuids(response.json()["cells"])
        assert uuids == [uuid_a, uuid_c, uuid_d, uuid_b]


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestUndoPermissions:
    def get_uuids(self, cells):
        def uuid(cell):
            return cell["metadata"]["jupyter_d1"]["uuid"]

        return list(map(uuid, cells))

    def test_undo_permission(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers)

        response = client.get(
            f"/notebooks/{uuid}/undo", headers=readonly_token_headers
        )
        assert response.status_code == 403
        response = client.get(
            f"/notebooks/{uuid}/undo", headers=superuser_token_headers
        )
        assert response.status_code == 204

    def test_redo_permission(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers)

        response = client.get(
            f"/notebooks/{uuid}/redo", headers=readonly_token_headers
        )
        assert response.status_code == 403
        response = client.get(
            f"/notebooks/{uuid}/redo", headers=superuser_token_headers
        )
        assert response.status_code == 204



from fastapi.testclient import TestClient

from jupyter_d1 import app

client = TestClient(app)


def test_hello():
    response = client.get("/hello")
    assert response.status_code == 200
    assert response.json() == {"message": "Hello World"}


def test_hello_there():
    response = client.get("/hello/there")
    assert response.status_code == 200
    assert response.json() == {"message": "Hello there"}
